#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ======================================================================
#
#  gensbo - a GENeral Simulation Based Optimizer
#
#           一个基于仿真的优化器
#
# ======================================================================
#
#  author        :    Mingtao Li
#  date          :    2019.03.26
#  contributors  :    Xiaohai Zhang
#  email         :    mingtao.li@gmail.com
#  url           :    https://www.mingtaoli.cn
#
# ======================================================================

from gensbo.core.variable import *
#from gensbo.core.objfunc import *
#from gensbo.core.constraint import *
from gensbo.core.problem import *

if __name__ == '__main__':
    print("Testing...")
    print("hello from core/__init__.py \n")
